import numpy as np
import paddle
import paddle.nn.functional as F
import ms_deform_attn as msda

        
if __name__ == '__main__':
    np.random.seed(114514)

    n = 2
    len_q = 100
    d_model = 256
    n_heads = 16
    n_levels = 3
    n_points = 10
    input_spatial_shapes = np.array([[256, 256], [128, 128], [64, 64]])
    len_in = (input_spatial_shapes[:, 0] * input_spatial_shapes[:, 1]).sum()
    im2col_step = 128

    value = np.random.rand(n, len_in, n_heads, d_model//n_heads)
    sampling_offsets = np.random.rand(n, len_q, n_heads, n_levels, n_points, 2)
    attention_weights = np.random.rand(n, len_q, n_heads, n_levels * n_points)
    attention_weights = np.exp(attention_weights) / np.exp(attention_weights).sum(axis=-1, keepdims=True)
    reference_points = np.random.rand(n, len_q, 1, n_levels, 1, 2)

    offset_normalizer = np.stack([input_spatial_shapes[..., 1], input_spatial_shapes[..., 0]], -1)
    sampling_locations = reference_points + sampling_offsets / offset_normalizer[None, None, None, :, None, :]

    value = paddle.to_tensor(value).astype('float32')
    value.stop_gradient = False
    input_spatial_shapes = paddle.to_tensor(input_spatial_shapes).astype('int64')
    sampling_locations = paddle.to_tensor(sampling_locations).astype('float32')
    sampling_locations.stop_gradient = False
    attention_weights = paddle.to_tensor(attention_weights).astype('float32')
    attention_weights.stop_gradient = False
    input_level_start_index = paddle.concat((paddle.zeros((1,), dtype='int64'), input_spatial_shapes.prod(1).cumsum(0)[:-1].astype('int64')))

    output = msda.ms_deform_attn(
            value, input_spatial_shapes, input_level_start_index, sampling_locations, attention_weights, im2col_step)
    print("Forward succedded!")
    output.mean().backward()
    print("Backward succedded!")
